import pandas as pd
from sklearn.compose import make_column_transformer
from sklearn.preprocessing import OneHotEncoder, StandardScaler
from sklearn.dummy import DummyClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import cross_validate
from sklearn.pipeline import make_pipeline


def classifier(train_df, target_col, numeric_feats = None, categorical_feats = None, cv = 5):
    """This function preprocess the data, fit baseline model(dummyclassifier) and logistic regression with default setups to provide data scientists 
        easy access to the common models results(scores). 
        
        Parameters
        ----------
        train_df : pandas.DataFrame
            The clean train data which includes target column.
        target_col : str
            The column of the train data that has the target values.
        numeric_feats = list
            The numeric features that needs to be considered in the model. If the user enters an empty list, the function will use all numeric columns.
        categorical_feats : list
            The categorical features that needs to be considered in the model. 
        cv : int, optional
            The number of folds on the data for train and validation set.
        Returns
        -------
        Data frame
            A data frame that includes test scores and train scores for each model.
        Examples
        -------
        >>> classifier(train_df, target_col = 'target', numerical_feats = [], categorical_features = [])
        >>> classifier(train_df, target_col = 'target', numeric_feats = ['danceability', 'loudness'], categorical_feats=['genre'], cv=10)
    """


    if (not(isinstance(train_df , pd.core.frame.DataFrame))):
        raise TypeError("Invalid function input. Please enter a data frame")
    if (not (train_df.isna().sum().sum() == 0)):
        raise ValueError("Invalid function input. Please pass a clean pandas data frame")
    if not(isinstance(numeric_feats , list)):
        raise TypeError("Numeric Features should be passed as a list")
    if not(isinstance(categorical_feats , list)):
        raise TypeError("Categorical Features should be passed as a list")
    if not(isinstance(target_col , str)):
        raise TypeError("Target column must be passed as a string")

    
    X_train = train_df.drop(columns=target_col, axis=1)
    y_train = train_df[target_col]


    if not isinstance(numeric_feats, list):
        raise TypeError("The numeric features have to be entered as a list")
    if not isinstance(categorical_feats , list):
        raise TypeError("The categorical features have to be entered as a list")
    
    if numeric_feats == None or numeric_feats==[]:
        numeric_feats = train_df.select_dtypes(include='number').columns.tolist()
    if categorical_feats == None or categorical_feats ==[]:
        categorical_feats = train_df.select_dtypes(exclude='number').columns.tolist()


    preprocessor = make_column_transformer(
        (StandardScaler(), numeric_feats),
        (OneHotEncoder(), categorical_feats))

    dummy = DummyClassifier()
    lr = make_pipeline(preprocessor, LogisticRegression())

    results = pd.Series(dtype='float64') 

    models = {"DummyClassifier": dummy, "LogisticRegression" : lr}

    for model in models :
        scores = cross_validate(models[model], X_train, y_train, return_train_score = True,cv = cv)
        mean_scores = pd.DataFrame(scores).mean().to_frame(model)
        results = pd.concat([results, mean_scores], axis = 1)
    results = results.drop(columns = 0, axis=1)
    
    return results







import pandas as pd
from sklearn.dummy import DummyRegressor
from sklearn.compose import make_column_transformer
from sklearn.model_selection import cross_validate, train_test_split
from sklearn.pipeline import Pipeline, make_pipeline
from sklearn.preprocessing import OneHotEncoder, OrdinalEncoder, StandardScaler
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.linear_model import LinearRegression, Ridge, RidgeCV

def regressor(train_df, target_col, numeric_feats = None, categorical_feats=None, cv=5):
    """This function preprocess the data, fit baseline model(dummyregresor) and ridge with default setups to provide data scientists 
        easy access to the common models results(scores). 

        Parameters
        ----------
        train_df : pandas.DataFrame
            The clean train data which includes target column.
        target_col : str
            The column of the train data that has the target values.
        numeric_feats = list, optional
            The numeric features that needs to be considered in the model. If the user do not define this argument, the function will assume all the columns except the identified ones in other arguments as numeric features.
        categorical_feats : list, optional
            The categorical columns for which needs onehotencoder preprocessing.  
        cv : int, optional
            The number of folds on the data for train and validation set.

        Returns
        -------
        Data frame
            A data frame that includes test scores and train scores for each model.
        Examples
        -------
        >>> regressor(train_df, target_col = 'popularity', categorical_features='genre')
        >>> regressor(train_df, target_col = 'popularity', numeric_feats = ['danceability', 'loudness'], categorical_feats=['genre'], cv=10)
    """
    

    #Checking the type of inputs
    if not(isinstance(train_df , pd.core.frame.DataFrame)):
        raise TypeError("train_df must be a pandas dataframe. Please pass a pd.core.frame.DataFrame train_df.")
    if not(isinstance(target_col , str)):
        raise TypeError("target_col must be a str. Please pass target column in str object.")
    if not(isinstance(numeric_feats , list)) and numeric_feats is not None:
        raise TypeError("numeric_feats must be a list. Please pass a list of numeric columns.")
    if categorical_feats is None :
        categorical_feats = []
    elif not(isinstance(categorical_feats , list)):
        raise TypeError("categorical_feats must be a list. Please pass a list of categorical columns.")

    #Checking the valid value for the inputs
    if (not (train_df.isna().sum().sum() == 0)) :
        raise ValueError(
            f"Invalid train_df input. Please pass a clean pandas data frame"
        )

    valid_target_col = train_df.select_dtypes('number').columns.tolist()

    if (not (target_col in valid_target_col)):
        raise ValueError(
            f"Invalid target_col input. Please use one out of {valid_target_col} in str"
        )

    X_train = train_df.drop(columns=target_col, axis=1)
    y_train = train_df[target_col]

    valid_numeric_col = X_train.select_dtypes('number').columns.tolist()
    valid_categorical_col =  X_train.select_dtypes(exclude='number').columns.tolist()

    if numeric_feats is None:
        numeric_feats = valid_numeric_col
    elif not (set(numeric_feats).issubset(set(valid_numeric_col))):
        raise ValueError(
            f"Invalid numeric features. Please use a sublist out of {valid_numeric_col}"
        )
    if not (set(categorical_feats).issubset(set(valid_categorical_col))):
        raise ValueError(
            f"Invalid categorical features. Please use a sublist out of {valid_categorical_col}"
        )

    
    #PReprocessing and modeling
    preprocessor = make_column_transformer(
        (StandardScaler(), numeric_feats),
        (OneHotEncoder(), categorical_feats)
    )

    dummy = DummyRegressor()
    ridge = make_pipeline(preprocessor, Ridge())
    ridge_cv = make_pipeline(preprocessor, RidgeCV(cv = cv))
    lr = make_pipeline(preprocessor, LinearRegression())

    results = pd.Series(dtype='float64') 

    models = {"DummyRegressor": dummy, "Ridge" : ridge, "RidgeCV" : ridge_cv, "linearRegression" : lr}

    for model in models :
        scores = cross_validate(models[model], X_train, y_train, return_train_score = True,cv = cv)
        mean_scores = pd.DataFrame(scores).mean().to_frame(model)
        results = pd.concat([results, mean_scores], axis = 1)
    results = results.drop(columns = 0, axis=1)
    
    return results


import pandas as pd
import altair as alt


import os
alt.renderers.enable('html')
# alt.renderers.enable('mimetype')


def plot_distributions(data, bins = 40, dist_cols=None, class_label=None):
    """This function creates numerical distribution plots on either all the numeric columns or the ones provided to it
        
        Parameters
        ----------
        data : pandas.DataFrame
            The dataframe for which distribution plot has to be created
        bins : int
            The number of bins for histogram plot
        dist_cols : list, optional
            The subset of numeric columns for which the histogram plots have to be generated
        class_label : str, optional
            The name of the target column only in case of classification dataset. For regression dataset, it is not required
        Returns
        -------
        chart_numeric
            The Altair object for the plot
        Examples
        -------
        >>> plot_distributions(data)
        >>> plot_distributions(data, dist_cols=['loudness', 'acousticness'], class_label='target')
    """
    if data is None:
        raise ValueError("Required arg 'data' cannot be empty")
    
    if not isinstance(data, pd.DataFrame):
        raise TypeError("Please enter data of type pd.DataFrame")

    if not isinstance(bins, int):
        raise TypeError("bins should be of type int")

    if dist_cols is not None:
        if not isinstance(dist_cols, list):
            raise TypeError("The entered dist_cols should be of 'list' type")
        else:
            numeric_features = dist_cols
    else:
        numeric_features = list(data.select_dtypes('number').columns)

    if class_label is None:

        chart_numeric = alt.Chart(data).mark_bar().encode(
        alt.X(alt.repeat(), type='quantitative', bin=alt.Bin(maxbins=bins)),
                y=alt.Y('count()')
                ).properties(
                    width=300,
                    height=200
                ).repeat(
                    numeric_features, columns = 4
                ) 
        
    else:
        if not type(class_label) == str:
            raise TypeError("`class_label` should be of string type")
        else:

            chart_numeric = alt.Chart(data).mark_bar(opacity=0.6).encode(
            alt.X(alt.repeat(), type='quantitative', bin=alt.Bin(maxbins=bins)),
                    y=alt.Y('count()', stack=False),
                    color=class_label+':N'
                    ).properties(
                        width=300,
                        height=200
                    ).repeat(
                        numeric_features, columns = 2
                    ) 
    return chart_numeric


def plot_corr(data, corr='spearman'):
    """This function creates correlation plot for all the columns in the dataframe
        
        Parameters
        ----------
        data : pandas.DataFrame
            The dataframe for which distribution plot has to be created
        corr : str
            The correlation method, which can be among 'spearman', 'kendall' or 'pearson'
            The default value is spearman
        Returns
        -------
        corr_plot
            The Altair object for the plot
        Examples
        -------
        >>> plot_corr(data)
        >>> plot_corr(data, corr='kendall')
    """    
    if data is None :
        raise ValueError("Required arg 'data' cannot be empty")
    if not isinstance(data, pd.DataFrame):
        raise TypeError("Please enter data of type pd.DataFrame")
    if corr == '':
        raise ValueError("'corr' cannot be empty")
    if corr not in ['spearman', 'pearson', 'kendall']:
        raise ValueError("corr should be one of these: 'spearman', 'pearson', 'kendall'")

    corr_df= data.corr(corr).stack().reset_index(name='corr')
    corr_plot = alt.Chart(corr_df, title='Correlation Plot among all features and target').mark_rect().encode(
        x=alt.X('level_0', title='All features'),
        y=alt.Y('level_1', title='All features'),
        tooltip='corr',
        color=alt.Color('corr', scale=alt.Scale(domain=(-1, 1), scheme='purpleorange')))
    return corr_plot
  

def plot_splom(data, pair_cols=None):
    """This function creates SPLOM plot for all the numeric columns in the dataframe or the ones passed by the user
        
        Parameters
        ----------
        data : pandas.DataFrame
            The dataframe for which distribution plot has to be created
        pair_cols : list
            The list of dataframe columns, for which correlation plot is to be generated
        Returns
        -------
        splom_chart
            The Altair object for the plot
        Examples
        -------
        >>> plot_splom(data)
        >>> plot_splom(data, pair_cols=['loudness', 'acousticness', 'energy'])
    """    
    if data is None :
        raise ValueError("Required arg 'data' cannot be empty")
    if not isinstance(data, pd.DataFrame):
        raise TypeError("Please enter data of type pd.DataFrame")
    if pair_cols==[]:
        raise ValueError("pair_cols should not be empty list")
    if pair_cols is None:
        pair_cols = list(data.select_dtypes('number').columns)
    else:
        if not isinstance(pair_cols, list):
            raise TypeError("The entered pair_cols should be of 'list' type")
    splom_chart = alt.Chart(data).mark_point(opacity=0.3, size=10).encode(
    alt.X(alt.repeat('row'), type='quantitative'),
    alt.Y(alt.repeat('column'), type='quantitative')
    ).properties(
        width=200,
        height=200
    ).repeat(
        column=pair_cols,
        row=pair_cols
    )
    return splom_chart
    



import pandas as pd
import numpy as np

def cleaner(input_dataframe,lower_case=False):
    """
        Load downloaded data, clean the dataset, remove the NA score. 
        Clean data(remove Nan rows, strip extra white spaces from column names, and data, convert all column names to lower case, etc)
        
        Parameters
        ----------
        input_dataframe: pandas.DataFrame
            Data set to clean
        drop_nans: boolean
             Drop all rows that have a NaN in any column (default: False)
        
        Returns
        --------
            (DataFrame):  A cleaned and simplified DataFrame of the relevant columns for summary and visualization,

        Examples
        --------
        >>>cleaner(example_data)
    """
    ## Handle dataframe type error (Check if dataframe is of type Pandas DataFrame)
    if not isinstance(input_dataframe, pd.DataFrame):

        raise TypeError(f"passed dataframe is of type {type(input_dataframe).__name__}, should be DataFrame")
        
    # Handle empty dataframe or dataframe with all NAN
    if (input_dataframe.empty or input_dataframe.dropna().empty):

        raise ValueError("passed dataframe is None")

    # Handle type Value error (Check if is a boolean')
    if not isinstance(lower_case, bool):

        raise TypeError("passed type should be a boolean")

    # Drop all rows that have a NaN in any column
    input_dataframe.dropna(inplace=True)
        
    # Strip extra white spaces from column names, and data
    input_dataframe = input_dataframe.rename(columns=lambda x: x.strip())
    
    # convert all column names to lower case (default: False)
    if lower_case:
        input_dataframe.columns= input_dataframe.columns.str.strip().str.lower()
        
    return input_dataframe
